/**
 * @name commander-assistant.js
 * @fileOverview The assistant for the Commander scene.
 */

/**
 * Constructor.
 * @param {CommanderAssistant} commanderAssistant
 */
function CommanderAssistant() {
    Mojo.Log.error("Constructor called...");
}

CommanderAssistant.Menu = {
	ClearCommandHistory: {
		label: $L('Clear Command History'),
		command: 'clear-commands-command'
	}
};

CommanderAssistant.prototype.setup = function() {
    Mojo.Log.error("Commander::setup called...");
	try {
		var appMenuModel = {
			visible: true,
			items: [
				CommanderAssistant.Menu.ClearCommandHistory
			]
		};

		this.controller.setupWidget(Mojo.Menu.appMenu, undefined, appMenuModel);

        this.commandInputTextAttributes = 
        {
            visible: true,
            multiline: false,
            textReplacement: false,
            focus: false,
			enterSubmits: true,
            requiresEnterKey: true,
            "changeOnKeyPress": true
        };
        this.commandInputTextModel = 
        {
            value: '',
            disabled: false
        };
        this.controller.setupWidget('commandInput', this.commandInputTextAttributes, this.commandInputTextModel);        
        this.commandInputElement = this.controller.get('commandInput');

        this.commandOutputTextAttributes = 
        {
            visible: true,
            multiline: true,
            focus: false,
            textFieldMode: 'sentence-case'
        };
        this.commandOutputTextModel = 
        {
            value: '',
            disabled: true
        };
        this.controller.setupWidget('commandOutput', this.commandOutputTextAttributes, this.commandOutputTextModel);
        this.commandOutputElement = this.controller.get('commandOutput');
		
        this.outputScrollerModel={
			mode: "vertical"
		};
		this.controller.setupWidget('outputScroller', this.outputScrollerModel);
		
        //this.commandInputTextChanged = this.commandInputTextChanged.bindAsEventListener(this);
        Mojo.Log.error("Commander::setup completed.");
	}
	catch (e) {
		Mojo.Log.logException(e, 'CommanderAssistant#setup');
	}


};

CommanderAssistant.prototype.cleanup = function() {
    Mojo.Log.error("Commander::cleanup called...");
};

CommanderAssistant.prototype.activate = function() {
    Mojo.Log.error("Commander::activate called...");
    //this.commandInputElement.observe(Mojo.Event.propertyChange, this.commandInputTextChanged);
	this.controller.listen(this.commandInputElement, Mojo.Event.propertyChange, this.execute.bindAsEventListener(this));
    Mojo.Log.error("Commander::activate finished...");
};

CommanderAssistant.prototype.deactivate = function() {
    Mojo.Log.error("Commander::deactivate called...");
    //this.commandInputElement.stopObserving(Mojo.Event.propertyChange, this.commandInputTextChanged);
//	this.commandInputElement.removeEventListener(Mojo.Event.keypress, this.execute);		
    Mojo.Log.error("Commander::deactivate finished...");
};


CommanderAssistant.prototype.commandInputTextChanged = function(event){

    var commandValue = event.value;
    Mojo.Log.error("Commander::commandInputTextChanged - %s", commandValue);

    this.commandText = commandValue;
};

CommanderAssistant.prototype.execute = function(event){
    Mojo.Log.error("Commander::Event Received");
    
    if(Mojo.Char.isEnterKey(event.originalEvent.keyCode)) {
            Mojo.Log.error("Commander::Event Enter");
            Mojo.Log.error("Commander::Event Text is: %s", this.commandText);
            if(this.commandText) {
                this.commandInputTextModel.disabled = true;
                this.controller.modelChanged(this.commandInputTextModel);

                this.controller.serviceRequest('palm://org.webosinternals.shell', {
                    method: 'runCmd',
                    parameters: {cmd: this.commandText},
                    onSuccess: this.onExecute.bind(this, true),
                    onFailure: this.onExecute.bind(this, false)
                });
                this.commandInputElement.focus();
            }
    } else {
        Mojo.Log.error("Commander::Event Text: %s", event.value);
        this.commandText = event.value;
    }
};

CommanderAssistant.prototype.onExecute = function(success, response) {


    if(success) {
        Mojo.Log.error("onExecute - success - %s", response.returnValue);
        this.commandOutputTextModel.value = "# " + this.commandText + "\n" + response.returnValue + "\n" + this.commandOutputTextModel.value;
        this.controller.modelChanged(this.commandOutputTextModel);  

    } else {
        Mojo.Log.error("onExecute - failure");
        this.commandOutputTextModel.value = "# " + this.commandText + "\nFailed.\n" + this.commandOutputTextModel.value;
        this.controller.modelChanged(this.commandOutputTextModel); 
    }
    this.commandInputTextModel.value = '';
    this.commandInputTextModel.disabled = false;
    this.controller.modelChanged(this.commandInputTextModel);
    this.commandText = "";
};
/**
 * handle a menu command.
 */
CommanderAssistant.prototype.handleCommand = function(event) {

	try {
		switch (event.type) {

			case Mojo.Event.command:

				switch (event.command) {

					case CommanderAssistant.Menu.ClearCommandHistory.command:
						this._clearCommandHistory();
						break;
                        
					case Mojo.Menu.prefsCmd:
						this._openPreferencesView();
						break;
				}
				break;

			case Mojo.Event.commandEnable:

				// Standard Application Menu commands.
				if (event.command === Mojo.Menu.helpCmd) {

					event.stopPropagation(); // Enable the chosen menuitems
					return;
				}

				if (event.command === MenuData.ApplicationMenu.ClearCommandHistory.command) {
					event.preventDefault();
					return;
				}

				break;
		}
	}
	catch (e) {
		Mojo.Log.logException(e, 'handleCommand');
	}
};

CommanderAssistant.prototype._openPreferencesView = function() {

	this.controller.stageController.pushScene('preferences');
};


CommanderAssistant.prototype._clearCommandHistory = function(){
    this.commandOutputTextModel.value = "";
    this.controller.modelChanged(this.commandOutputTextModel); 
};

CommanderAssistant.prototype._showHelpCommand = function() {

	AppAssistant.launchHelp();
};

