/**
 * helpers for scene assistants in Spaz on Mojo 
 */
var scene_helpers = {}
 
 
/**
 * This adds a number of common scene methods to the passed scene assistant
 * @param {object} assistant a scene assistant
 */
scene_helpers.addCommonSceneMethods = function(assistant) {

/**
   * This helps us set up listening for the Enter key in a textbox
   * 
   * the callback function's 'this' references the assistant 
   * 
   * make sure to call stopListeningForEnter when done with the
   * correct ID so the listener is unbound
   * 
   * @param {string} id
   * @param {function} callback
   */
  assistant.listenForEnter = function (id, callback){
    Mojo.Event.listen(this.controller.get(id),
      Mojo.Event.propertyChange,
      this._listenerForEnter.bind(this, callback),
      true
    );
  }
  assistant._listenerForEnter = function(callback, event) {
    if (event && Mojo.Char.isEnterKey(event.originalEvent.keyCode)) {
      callback.call(this);
      return;
    }
  };
  
  /**
   * removes the listener set up by listenForEnter
   * 
   * @param {string} id
   */
  assistant.stopListeningForEnter = function(id) {
    Mojo.Event.stopListening(this.controller.get(id),
      Mojo.Event.propertyChange,
      this._listenerForEnter
    );
  }
});