/*
 * Copyright (c) 2009 by the authors listed in the AUTHORS file.
 * MIT License
 */
package org.webosinternals.services;

import com.palm.luna.LSException;
import com.palm.luna.service.LunaServiceThread;
import com.palm.luna.service.ServiceMessage;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import org.json.JSONException;
import org.json.JSONObject;

/**
 *
 * @author Brian Jost aka PreGame
 * Credit to ShinAli for showing the basis of creating a service for WebOS
 */
public class ShellService extends LunaServiceThread {

    /**
     * Description: Run a raw command and send back the command output/return value.
     *
     * @param msg - Command that the user wants to run in ServiceMessage format
     * @throws org.json.JSONException
     * @throws java.io.IOException
     * @throws com.palm.luna.LSException
     *
     * Respond: Sends back command output
     */
    @LunaServiceThread.PublicMethod
    public void runCmd(ServiceMessage msg) throws JSONException, IOException, LSException
    {

	//Check if the cmd param was specified
	if (msg.getJSONPayload().has("cmd"))
	{
	    String cmd = msg.getJSONPayload().getString("cmd");
	    // Setup the process to execute the command.
	    Process p = Runtime.getRuntime().exec(cmd);
	    BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
	    String line = null;
	    String returnValue = "";
	    while ((line = input.readLine()) != null)
	    {
		returnValue += line;
		// the below line is used to somewhat give the effect of an actual terminal output
		returnValue += System.getProperty("line.separator");
	    }

	    input.close();
	    JSONObject reply = new JSONObject();
	    reply.put("returnValue", returnValue);
	    msg.respond(reply.toString());
	}
	else
	{
	    msg.respondError("ERROR", "Please provide a command to run.");
	}
    }
}
