/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Josh Aas <josh@mozilla.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <sys/time.h>

#include "termplugin.h"
#include "nppalmdefs.h"
#include "pixelbuffer.hpp"
#include "charbuffer.hpp"
#include "spawn.h"
#include "keyman.hpp"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <algorithm>

using namespace std;

#define PLUGIN_NAME        "Term Plug-in"
#define PLUGIN_DESCRIPTION PLUGIN_NAME " (Mozilla SDK)"
#define PLUGIN_VERSION     "1.0.0.0"
#define PLUGIN_MIMEDESCRIPTION "application/x-webosinternals-termplugin:trm:Term plugin"

static NPNetscapeFuncs NPNFuncs;


static double CurrentTime()
{
        struct timeval tv;
        gettimeofday(&tv, 0);
        return tv.tv_sec + tv.tv_usec/1000000.0;
}


void logTimeStamp(FILE *out, int line)
	{
	struct tm *t;
	time_t now;

	if(!out)
		return;
	time(&now);
	t = localtime(&now);

	fprintf(out, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d	%d	term: ",
			t->tm_year + 1900,
			t->tm_mon + 1,
			t->tm_mday,
			t->tm_hour,
			t->tm_min,
			t->tm_sec, line);
	}

#define debugLogLine(x)								debugLog(__LINE__, x)
#if 0
static void debugLogArgv(int line, int argc, char **argn, char **argv, const char *msg)
	{
	int j;
	const char *prefix = "args =";
#if !defined(__STDERR_NO_GOOD__)
	FILE *out = stderr;
#else
	FILE *out = fopen("/tmp/termPluginOutput.log", "a");
	if(!out)
		return;
#endif
	logTimeStamp(out, line);
	fprintf(out, "%s(", msg);
	for(j = 0; argc > 0; argc--, argv++, argn++, j++, prefix = ",")
		fprintf(out, "%s %d:%s[%s]", prefix, j, *argn, *argv);
	fprintf(out, ")\n");
#if defined(__STDERR_NO_GOOD__)
	if(out != stderr)
		fclose(out);
#endif
	}
#endif
#include <stdarg.h>
void debugLog(int line, const char *format, ...)
	{
#if !defined(__STDERR_NO_GOOD__)
	FILE *out = stderr;
#else
	FILE *out = fopen("/tmp/termPluginOutput.log", "a");
	if(!out)
		return;
#endif
	logTimeStamp(out, line);
	va_list(arglist);
	va_start(arglist, format);
	vfprintf(out, format, arglist);
	fprintf(out, "\n");
#if defined(__STDERR_NO_GOOD__)
	if(out != stderr)
		fclose(out);
#endif
	}
#if 0
void debugLogVariant(int line, NPVariant *var, bool ok, const char *msg)
	{
	if(ok)
		{
		if(NPVARIANT_IS_INT32(*var))
			debugLog(line, "%s = %d val:%d", msg, ok, NPVARIANT_TO_INT32(*var));
		else if(NPVARIANT_IS_DOUBLE(*var))
			debugLog(line, "%s = %d val:%f", msg, ok, NPVARIANT_TO_DOUBLE(*var));
		else if(NPVARIANT_IS_BOOLEAN(*var))
			debugLog(line, "%s = %d val:%s", msg, ok, NPVARIANT_TO_BOOLEAN(*var) ? "true" : "false");
		else if(NPVARIANT_IS_STRING(*var))
			debugLog(line, "%s = %d val:[%s]", msg, ok, NPVARIANT_TO_STRING(*var).UTF8Characters);
		else if(NPVARIANT_IS_OBJECT(*var))
			debugLog(line, "%s = %d val:%p", msg, ok, NPVARIANT_TO_OBJECT(*var));
		else
			debugLog(line, "%s = %d type:%d", msg, ok, var->type);
		}
	else
		debugLog(line, "%s = %d", msg, ok);
	}
#endif










struct MyObject : NPObject, KeyManager::Client {
  NPP instance;
  KeyManager key_manager;
  int event_x;
  int event_y;
  int terminalHeight;
  typedef CharBuffer CharBuf;
  CharBuf char_buf;
  SpawnData *spawn_data;
  pthread_t select_thread;
  bool select_thread_started;
  pthread_mutex_t draw_mutex;
  pthread_cond_t drawn_cond;

  MyObject(NPP i);
  ~MyObject();

  void checkSpawn();

  void sendChars(const char *buf,int len)
  {
    {
      fprintf(stderr,"Sending");
      int i = 0;
      for (;i<len;++i) {
        fprintf(stderr," %d",buf[i]);
      }
      fprintf(stderr,"\n");
    }
    Spawn_Send(spawn_data,buf,len);
  }

  static const int screen_width = 480 / 8;		//	since landscape_width is 480, this will avoid the realloc when user flips screen the 1st time (the black/white screen)
  static const int screen_height = 400 / 8;		//	since the JS has normal_height as 400, let the default here be the same!
  static const int scroll_height = 1000;		//	since the JS is using 1000 lines...
  static CommandHandlers term_handlers;

  static int _termVoidOp(void *obj,TermOp_t op)
  {
    ((MyObject *)obj)->_termVoid(op);
    return 1;
  }

  static int _termIntOp(void *obj,TermOp_t op,int arg1)
  {
    ((MyObject *)obj)->_termInt(op,arg1);
    return 1;
  }

  static int _termIntIntOp(void *obj,TermOp_t op,int arg1,int arg2)
  {
    ((MyObject *)obj)->_termIntInt(op,arg1,arg2);
    return 1;
  }

  static int _termStringOp(void *,TermOp_t op,const char *arg)
  {
    fprintf(stderr,"term: string op\n");
    return 1;
  }

  static void _termTextOp(void *obj_ptr,const char *buf,int n)
  {
    ((MyObject *)obj_ptr)->_termText(buf,n);
  }

  static void _termUnexpected(void *,int)
  {
    fprintf(stderr,"term: unexpected\n");
  }

  void _termVoid(TermOp_t op);
  void _termInt(TermOp_t op,int arg1);
  void _termIntInt(TermOp_t op,int arg1,int arg2);
  void _termText(const char *buf,int n);
  void startSelectThread();
  void stopSelectThread();
  static void *selectThread(void *);
  void runSelect();
  static void masterReady(void *obj_ptr);
};

MyObject::MyObject(NPP i) 
: instance(i),
  key_manager(this),
  event_x(0),
  event_y(0),
  char_buf(screen_width,scroll_height,screen_width,screen_height),
  spawn_data(0),
  select_thread_started(false)
{
  term_handlers.void_proc = MyObject::_termVoidOp;
  term_handlers.int_proc = MyObject::_termIntOp;
  term_handlers.int_int_proc = MyObject::_termIntIntOp;
  term_handlers.string_proc = MyObject::_termStringOp;
  term_handlers.text_proc = MyObject::_termTextOp;
  term_handlers.unexpected_proc = MyObject::_termUnexpected;
#if 0
  {
    close(2);
    int fd = creat("/tmp/term.out",S_IRUSR|S_IWUSR);
    if (fd!=2) {
      dup2(fd,2);
      close(fd);
    }
  }
#endif
  spawn_data = Spawn_Open(&term_handlers,this);
  if (spawn_data) {
    const char *argv[] = {"/bin/login",0};
    Spawn_Stty(spawn_data,screen_height,screen_width);
    Spawn_Exec(spawn_data,(char **)argv);
    startSelectThread();
  }
}

MyObject::~MyObject()
{
  if (select_thread_started) {
    stopSelectThread();
  }
  Spawn_Kill(spawn_data,"INT");
  Spawn_CheckChildExit(spawn_data);
  Spawn_Delete(spawn_data);
  spawn_data = 0;
}

void MyObject::masterReady(void *obj_ptr)
{
  fprintf(stderr,"term: masterReady: forcing redraw\n");
  NPN_ForceRedraw(((MyObject *)obj_ptr)->instance);
}

void MyObject::runSelect()
{
  fprintf(stderr,"In select thread.\n");
  for (;;) {
    Spawn_WaitMaster(spawn_data);
#if 0
    fprintf(stderr,"term: calling masterReady\n");
    fprintf(stderr,"term: func=%p\n",NPNFuncs.pluginthreadasynccall);
    fprintf(stderr,"term: instance=%p\n",instance);
    NPNFuncs.pluginthreadasynccall(instance,MyObject::masterReady,this);
#else
    pthread_mutex_lock(&draw_mutex);
    fprintf(stderr,"Forcing redraw\n");
    NPN_ForceRedraw(instance);
    fprintf(stderr,"Waiting for redraw\n");
    pthread_cond_wait(&drawn_cond,&draw_mutex);
    fprintf(stderr,"Done redraw\n");
    pthread_mutex_unlock(&draw_mutex);
#endif
  }
}

void *MyObject::selectThread(void *obj_ptr)
{
  ((MyObject *)obj_ptr)->runSelect();
  return 0;
}

void MyObject::startSelectThread()
{
  pthread_mutex_init(&draw_mutex,NULL);
  pthread_cond_init(&drawn_cond,NULL);
  int rc = pthread_create(&select_thread,NULL,MyObject::selectThread,this);

  if (rc!=0) {
    fprintf(stderr,"Failed to create select thread.\n");
  }
  else {
    fprintf(stderr,"Created select thread\n");
    select_thread_started = true;
  }
}


void MyObject::stopSelectThread()
{
  pthread_cancel(select_thread);
  void *thread_return = 0;
  pthread_join(select_thread,&thread_return);
  fprintf(stderr,"Select thread terminated\n");
  select_thread_started = false;
}


void MyObject::_termText(const char *buf,int n)
{
  {
    fprintf(stderr,"_termText: '");
    int i = 0;
    for (;i!=n;++i) {
      fprintf(stderr,"%c",buf[i]);
    }
    fprintf(stderr,"'\n");
  }
  int i = 0;
  for (;i!=n;++i) {
    char_buf.put(buf[i]);
    char_buf.advanceCursor();
    //fprintf(stderr,"cursor at %d,%d\n",char_buf.cursor().row(),char_buf.cursor().col());
  }
}


void MyObject::_termVoid(TermOp_t op)
{
  fprintf(stderr,"_termVoid: %s\n",TermState_OpName(op));
  switch (op) {
    case TERM_HOME_CURSOR:
      char_buf.homeCursor();
      break;
    case TERM_CLEAR_BELOW:
      char_buf.clearToRight();
      char_buf.clearBelow();
      break;
    case TERM_CLEAR_ABOVE:
      break;
    case TERM_CLEAR_TO_RIGHT:
      char_buf.clearToRight();
      break;
    case TERM_CLEAR_TO_LEFT:
      break;
    case TERM_CLEAR_SCREEN:
      break;
    case TERM_CLEAR_LINE:
      break;
    case TERM_NORM_ATTR:
      char_buf.setFgColor(2);		//	Green - these should be defined somewhere
      char_buf.setBgColor(0);		//	Black - these should be defined somewhere
      char_buf.setFlag(eFlagsAll, 0);
      break;
    case TERM_SAVE_CURSOR:
      break;
    case TERM_RESTORE_CURSOR:
      break;
    case TERM_BELL:
      break;
    case TERM_BACKSPACE:
      char_buf.backSpaceCursor();
      break;
    case TERM_TAB:
      break;
    case TERM_CLEAR_TABS:
      break;
    case TERM_CARRIAGE_RETURN:
      char_buf.carriageReturn();
      break;
    case TERM_MOVE_UP:
      break;
    case TERM_MOVE_DOWN:
      char_buf.moveCursorDown();
      break;
    case TERM_EOF:
      break;
    case TERM_BUFFER_EMPTY:
      break;
    case TERM_REQ_ATTR:
      break;
    case TERM_PRIMARY_FONT:
      break;
    case TERM_ALTERNATE_FONT1:
      break;
    case TERM_ALTERNATE_FONT2:
      break;
    default:
      assert(0);
  }
}


void MyObject::_termInt(TermOp_t op,int arg1)
{
  fprintf(stderr,"_termInt: %s %d\n",TermState_OpName(op),arg1);
  switch (op) {
    case TERM_CURSOR_UP:
      char_buf.moveCursorUp(arg1);
      break;
    case TERM_CURSOR_DOWN:
      char_buf.moveCursorDown(arg1);
      break;
    case TERM_CURSOR_RIGHT:
      char_buf.moveCursorRight(arg1);
      break;
    case TERM_CURSOR_LEFT:
      char_buf.moveCursorLeft(arg1);
      break;
    case TERM_INSERT_LINE:
      break;
    case TERM_DELETE_LINE:
      break;
    case TERM_INSERT_CHAR:
      break;
    case TERM_ERASE_CHAR:
      break;
    case TERM_DELETE_CHAR:
      break;
    case TERM_FGCOLOR_ATTR:
      char_buf.setFgColor(arg1);
      break;
    case TERM_BGCOLOR_ATTR:
      char_buf.setBgColor(arg1);
      break;
    case TERM_BOLD_ATTR:
      char_buf.setFlag(eBold, arg1);
      break;
    case TERM_DIM_ATTR:
      char_buf.setFlag(eDim, arg1);
      break;
    case TERM_ITALIC_ATTR:
      char_buf.setFlag(eItalics, arg1);
      break;
    case TERM_UNDERLINE_ATTR:
      char_buf.setFlag(eUnderline, arg1);
      break;
    case TERM_BLINK_ATTR:
      char_buf.setFlag(eBlink, arg1);
      break;
    case TERM_INVERSE_ATTR:
      char_buf.setFlag(eInverse, arg1);
      break;
    case TERM_APP_CURSOR_KEYS:
      key_manager.setAppCursorKeys(arg1);
      break;
    default:
      break;
  }
}


void MyObject::_termIntInt(TermOp_t op,int arg1,int arg2)
{
  fprintf(stderr,"_termIntInt: %s %d %d\n",TermState_OpName(op),arg1,arg2);
  switch (op) {
    case TERM_SET_CURSOR:
      char_buf.setCursor(arg1,arg2);
      break;
    default:
      break;
  }
}


CommandHandlers MyObject::term_handlers;

void MyObject::checkSpawn()
{
  if (spawn_data) {
    if (!Spawn_ReadMaster(spawn_data)) {
      spawn_data = 0;
    }
  }
}

struct InstanceData {
//  NPP npp;
//  NPWindow window;
  MyObject *object;
  InstanceData() : object(0) 
  { 
  }
};

void logmsgv(const char* desc, NPPVariable variable) 
{
    fprintf(stderr, "term: %s - %s (#%i)\n", desc, varname(variable), variable);
}

static void
fillPluginFunctionTable(NPPluginFuncs* pFuncs)
{
  pFuncs->version = 11;
  pFuncs->size = sizeof(*pFuncs);
  pFuncs->newp = NPP_New;
  pFuncs->destroy = NPP_Destroy;
  pFuncs->setwindow = NPP_SetWindow;
  pFuncs->newstream = NPP_NewStream;
  pFuncs->destroystream = NPP_DestroyStream;
  pFuncs->asfile = NPP_StreamAsFile;
  pFuncs->writeready = NPP_WriteReady;
  pFuncs->write = NPP_Write;
  pFuncs->print = NPP_Print;
  pFuncs->event = NPP_HandleEvent;
  pFuncs->urlnotify = NPP_URLNotify;
  pFuncs->getvalue = NPP_GetValue;
  pFuncs->setvalue = NPP_SetValue;
}

NP_EXPORT(NPError)
NP_Initialize(NPNetscapeFuncs* bFuncs, NPPluginFuncs* pFuncs)
{
  fprintf(stderr,"term: NP_Initialize\n");
  fprintf(stderr,"term: bFuncs->size=%d\n",bFuncs->size);
  fprintf(stderr,"term: sizeof(NPNetscapeFuncs)=%d\n",sizeof(NPNetscapeFuncs));

  NPNFuncs = *bFuncs;

  fillPluginFunctionTable(pFuncs);

  return NPERR_NO_ERROR;
}

NP_EXPORT(char*)
NP_GetPluginVersion()
{
  fprintf(stderr, "term: NP_GetPluginVersion\n");
  return (char *)PLUGIN_VERSION;
}

NP_EXPORT(char*)
NP_GetMIMEDescription()
{
  fprintf(stderr,"term: NP_GetMIMEDescription\n");
  return (char *)PLUGIN_MIMEDESCRIPTION;
}

NPObject *NPN_CreateObject(NPP npp, NPClass *aClass)
{
  return NPNFuncs.createobject(npp, aClass);
}

void NPN_ForceRedraw(NPP npp)
{
  return NPNFuncs.forceredraw(npp);
}

bool NPN_IdentifierIsString(NPIdentifier identifier)
{
  return NPNFuncs.identifierisstring(identifier);
}

NPUTF8 *NPN_UTF8FromIdentifier(NPIdentifier identifier)
{
  return NPNFuncs.utf8fromidentifier(identifier);
}

static NPObject *Object_Allocate(NPP npp,NPClass *aClass)
{
  return new MyObject(npp);
}

static void Object_Deallocate(NPObject *npobj)
{
  fprintf(stderr,"term: In Object_Deallocate\n");
  delete (MyObject *)npobj;
}

static bool 
  Object_Construct(
    NPObject *npobj, 
    const NPVariant *args,
    uint32_t argCount,
    NPVariant *result
  )
{
  fprintf(stderr,"In Object_Construct\n");
  return false;
}

static bool ContainsName(const char **name_array,NPIdentifier name)
{
  bool is_string = NPN_IdentifierIsString(name);
  if (!is_string) {
    fprintf(stderr,"  Identifier is not a string.\n");
    return false;
  }
  const char *name_str = NPN_UTF8FromIdentifier(name);
  int index = 0;
  for (;;++index) {
    const char *one_name = name_array[index];
    if (!one_name) {
      break;
    }
    if (strcmp(name_str,one_name)==0) {
      return true;
    }
  }
  return false;
}


static bool Object_HasProperty(NPObject *npobj, NPIdentifier name)
{
  static const char *property_names[] = {
    0
  };
  return ContainsName(property_names,name);
}


static bool Object_HasMethod(NPObject *npobj,NPIdentifier name)
{
  static const char *method_names[] = {
    "initialize",
    "sendTap",
    "sendMouseDown",
    "sendMouseUp",
    "sendKeyDown",
    "sendKeyUp",
    "sendKeyPress",
    "setTerminalHeight",
    "sendGestureStart",
    "sendGestureEnd",
    "sendGestureChange",
    "test",
    0
  };
  return ContainsName(method_names,name);
}


static void PrintValue(const NPVariant *arg)
{
  if (NPVARIANT_IS_VOID(*arg)) {
    fprintf(stderr,"void");
  }
  else if (NPVARIANT_IS_NULL(*arg)) {
    fprintf(stderr,"null");
  }
  else if (NPVARIANT_IS_BOOLEAN(*arg)) {
    if (NPVARIANT_TO_BOOLEAN(*arg)) {
      fprintf(stderr,"true");
    }
    else {
      fprintf(stderr,"false");
    }
  }
  else if (NPVARIANT_IS_INT32(*arg)) {
    fprintf(stderr,"%d",NPVARIANT_TO_INT32(*arg));
  }
  else if (NPVARIANT_IS_DOUBLE(*arg)) {
    fprintf(stderr,"%g",NPVARIANT_TO_DOUBLE(*arg));
  }
  else if (NPVARIANT_IS_STRING(*arg)) {
    NPString value = NPVARIANT_TO_STRING(*arg);
    const NPUTF8 *chars = value.UTF8Characters;
    uint32_t len = value.UTF8Length;
    fprintf(stderr,"\"");
    {
      uint32_t i = 0;
      for (;i!=len;++i) {
        fprintf(stderr,"%c",chars[i]);
      }
    }
    fprintf(stderr,"\"");
  }
  else if (NPVARIANT_IS_OBJECT(*arg)) {
    fprintf(stderr,"object");
  }
  else {
    fprintf(stderr,"unknown");
  }
}


static void PrintArgs(const NPVariant *args,int argCount)
{
  fprintf(stderr,"(");
  int argnum = 0;
  for (;argnum!=argCount;++argnum) {
    if (argnum!=0) {
      fprintf(stderr,",");
    }
    PrintValue(&args[argnum]);
  }
  fprintf(stderr,")");
}


static int IntValue(const NPVariant *value)
{
  if (NPVARIANT_IS_INT32(*value)) {
    int32_t value_int32 = NPVARIANT_TO_INT32(*value);
    return value_int32;
  }
  else if (NPVARIANT_IS_DOUBLE(*value)) {
    double value_double = NPVARIANT_TO_DOUBLE(*value);
    return (int)value_double;
  }
  else if (NPVARIANT_IS_BOOLEAN(*value)) {
    bool value_bool = NPVARIANT_TO_BOOLEAN(*value);
    return (int)value_bool;
  }
  else if (NPVARIANT_IS_STRING(*value)) {
    const NPString &value_str = NPVARIANT_TO_STRING(*value);
    const char *value_chars = value_str.UTF8Characters;
    fprintf(stderr,"String is %s\n",value_chars);
    return (int)value_chars[0];
  }
  else {
    fprintf(stderr,"Expecting int, but got type %d\n",value->type);
    return 0;
  }
}


static bool BoolValue(const NPVariant *value)
{
  if (NPVARIANT_IS_BOOLEAN(*value)) {
    bool value_bool = NPVARIANT_TO_BOOLEAN(*value);
    return value_bool;
  }
  else {
    return IntValue(value)!=0;
  }
}


bool 
  Object_Invoke(
    NPObject *npobj,
    NPIdentifier name,
    const NPVariant *args,
    uint32_t argCount,
    NPVariant *result
  )
{
  MyObject *myobj_ptr = (MyObject *)npobj;
  assert(myobj_ptr);
  MyObject &myobj = *myobj_ptr;
  bool is_string = NPN_IdentifierIsString(name);
  if (!is_string) {
    fprintf(stderr,"  Object_Invoke: Method name is not a string\n");
    return false;
  }
  const char *name_str = NPN_UTF8FromIdentifier(name);
  fprintf(stderr,"  Object_Invoke: %s",name_str);
  PrintArgs(args,argCount);
  fprintf(stderr,"\n");
  if (strcmp(name_str,"sendTap")==0) {
    if (argCount<2) {
      fprintf(stderr,"  wrong number of args\n");
      return false;
    }
    myobj.event_x = IntValue(&args[0]);
    myobj.event_y = IntValue(&args[1]);
    NPN_ForceRedraw(myobj.instance);
    return true;
  }
  else if (strcmp(name_str,"setTerminalHeight")==0/* || strcmp(name_str,"initialize")==0*/) {
    if (argCount<1) {
      fprintf(stderr,"  wrong number of args\n");
      return false;
    }
    int terminalHeight = IntValue(&args[0]);
    fprintf(stderr,"  setTerminalHeight(%d)\n",terminalHeight);
    int char_height = myobj.char_buf.font().char_height();
    myobj.char_buf.setVisibleHeight(terminalHeight/char_height);
    return true;
  }
  else if (strcmp(name_str,"sendKeyDown")==0) {
    if (argCount<5) {
      fprintf(stderr,"  wrong number of args\n");
      return false;
    }
    int key_code = IntValue(&args[0]);
    bool shift_pressed = BoolValue(&args[1]);
    bool alt_pressed = BoolValue(&args[2]);
    bool ctrl_pressed = BoolValue(&args[3]);
    bool meta_pressed = BoolValue(&args[4]);

    myobj.key_manager.keyPressed(
      key_code,shift_pressed,alt_pressed,ctrl_pressed,meta_pressed
    );
    return true;
  }

  return false;
}


static bool 
  Object_SetProperty(
    NPObject *npobj, 
    NPIdentifier name,
    const NPVariant *value
  )
{
  MyObject *myobj = (MyObject *)npobj;
  fprintf(stderr,"In Object_SetProperty\n");
  bool is_string = NPN_IdentifierIsString(name);
  fprintf(stderr,"  IdentifierIsString = %d\n",is_string);
  if (is_string) {
    const char *name_str = NPN_UTF8FromIdentifier(name);
    fprintf(stderr,"  Identifier = %s\n",name_str);
    if (strcmp(name_str,"event_x")==0) {
      myobj->event_x = IntValue(value);
      fprintf(stderr,"  event_x set to %d\n",myobj->event_x);
    }
    else if (strcmp(name_str,"event_y")==0) {
      myobj->event_y = IntValue(value);
      fprintf(stderr,"  event_y set to %d\n",myobj->event_y);
    }
    else if (strcmp(name_str,"test_property")==0) {
      int command = IntValue(value);
      fprintf(stderr,"  command=%d\n",command);
      if (command==2) {
        NPN_ForceRedraw(myobj->instance);
      }
    }
    else {
      fprintf(stderr,"Invalid property\n");
    }
  }
  return false;
}

static NPObject *GetScriptableObject(NPP instance)
{
  fprintf(stderr,"GetScriptableObject: instance=%p\n",instance);
  InstanceData *data = (InstanceData *)instance->pdata;
  if (!data->object) {
    static NPClass object_class;
    object_class.structVersion = NP_CLASS_STRUCT_VERSION;
    object_class.allocate = Object_Allocate;
    object_class.deallocate = Object_Deallocate;
    object_class.invalidate = 0;
    object_class.hasMethod = Object_HasMethod;
    object_class.invoke = Object_Invoke;
    object_class.invokeDefault = 0;
    object_class.hasProperty = Object_HasProperty;
    object_class.getProperty = 0;
    object_class.setProperty = Object_SetProperty;
    object_class.removeProperty = 0;
    object_class.enumerate = 0;
    object_class.construct = Object_Construct;

    data->object = 
      (MyObject *)NPN_CreateObject(
        instance,
        &object_class
      );
  }

  return data->object;
}

NP_EXPORT(NPError)
NP_GetValue(NPP instance, NPPVariable aVariable, void* aValue) {

    switch (aVariable) {
    case NPPVpluginNameString:
      *((const char**)aValue) = PLUGIN_NAME;
      break;
    case NPPVpluginDescriptionString:
      *((const char**)aValue) = PLUGIN_DESCRIPTION;
      break;
    default:
      return NPERR_INVALID_PARAM;
      break;
  }
  return NPERR_NO_ERROR;
}

NP_EXPORT(NPError)
NP_Shutdown()
{
  fprintf(stderr,"term: NP_Shutdown\n");
  return NPERR_NO_ERROR;
}

NPError
NPP_New(NPMIMEType pluginType, NPP instance, uint16_t mode, int16_t argc, char* argn[], char* argv[], NPSavedData* saved) {
  instance->pdata = new InstanceData;
  fprintf(stderr,"term: NP_New\n");
  return NPERR_NO_ERROR;
}

NPError
NPP_Destroy(NPP instance, NPSavedData** save) {
  fprintf(stderr,"term: NPP_Destroy\n");
  InstanceData* instanceData = (InstanceData*)(instance->pdata);
  if (instanceData) {
    delete instanceData;
  }
  return NPERR_NO_ERROR;
}


NPError
NPP_SetWindow(NPP instance, NPWindow* window) {
        if (!window) {
                // Destroying window
		debugLog(__LINE__, "NPP_SetWindow(window:%p)", window);
                return NPERR_NO_ERROR;
                }
        if (!window->window) {
		debugLog(__LINE__, "NPP_SetWindow(window->window:%p)", window->window);
                return NPERR_NO_ERROR;
                }
#if 0
        debugLog(__LINE__, "NPP_SetWindow(x:%d, y:%d, width:%d, height:%d, type:%d)", window->x, window->y, window->width, window->height, window->type);
#endif

        InstanceData *data = (InstanceData *)instance->pdata;
        MyObject *myobj_ptr = data->object;
        if (!myobj_ptr) {
                return NPERR_NO_ERROR;
                }
        CharBuffer &char_buf = myobj_ptr->char_buf;
        if (!char_buf.resizeToPixels(window->width, window->height)) {
                return NPERR_NO_ERROR;
                }
        fprintf(stderr,"term: visibile size changed.\n");
        Spawn_Stty(
                myobj_ptr->spawn_data,
                char_buf.visibleHeight(),
                char_buf.visibleWidth()
        );
        return NPERR_NO_ERROR;
}


NPError
NPP_NewStream(NPP instance, NPMIMEType type, NPStream* stream, NPBool seekable, uint16_t* stype) {
  fprintf(stderr,"term: NPP_NewStream\n");
  return NPERR_GENERIC_ERROR;
}

NPError
NPP_DestroyStream(NPP instance, NPStream* stream, NPReason reason) {
  fprintf(stderr,"term: NPP_DestroyStream\n");
  return NPERR_GENERIC_ERROR;
}

int32_t
NPP_WriteReady(NPP instance, NPStream* stream) {
  fprintf(stderr,"term: NPP_WriteReady\n");
  return 0;
}

int32_t
NPP_Write(NPP instance, NPStream* stream, int32_t offset, int32_t len, void* buffer) {
  fprintf(stderr,"term: NPP_Write\n");
  return 0;
}

void
NPP_StreamAsFile(NPP instance, NPStream* stream, const char* fname) {
  fprintf(stderr,"term: NPP_StreamAsFile\n");
}

void
NPP_Print(NPP instance, NPPrint* platformPrint) {
  fprintf(stderr,"term: NPP_Print\n");
}


void HandlePenMoveEvent(NpPalmPenEvent *penEvent,NPP instance)
{
  fprintf(stderr,"HandlePenMoveEvent: x=%d, y=%d\n",penEvent->xCoord,penEvent->yCoord);
}


void HandleKeyEvent(NpPalmKeyEvent *keyEvent,NPP instance)
{
  int key_code = keyEvent->rawkeyCode;
  int modifiers = keyEvent->rawModifier;

  fprintf(
    stderr,
    "term: keyEvent key=%d, modifiers=%d\n",
    key_code,
    modifiers
  );
#if 1
  InstanceData *data = (InstanceData *)instance->pdata;
  MyObject *myobj_ptr = data->object;
  if (!myobj_ptr) {
    return;
  }
  MyObject &myobj = *myobj_ptr;

  bool shift_pressed = (modifiers & 128)!=0;
  bool ctrl_pressed = (modifiers & 64)!=0;
  bool alt_pressed = false;
  bool meta_pressed = (modifiers & 32)!=0;

  myobj.key_manager.keyPressed(
    key_code,shift_pressed,alt_pressed,ctrl_pressed,meta_pressed
  );
#endif
}

void HandleDrawEvent(NpPalmDrawEvent *drawEvent,NPP instance)
{
  fprintf(stderr,"term: drawEvent sl: %i sr: %i st: %i sb: %i dl: %i dr: %i dt: %i db: %i drb:%i\n", 
    drawEvent->srcLeft, 
    drawEvent->srcRight, 
    drawEvent->srcTop, 
    drawEvent->srcBottom, 
    drawEvent->dstLeft, 
    drawEvent->dstRight, 
    drawEvent->dstTop, 
    drawEvent->dstBottom,
    drawEvent->dstRowBytes
  ); 
  double start_time = CurrentTime();
  // PGContextIface *iface = (PGContextIface *) drawEvent->graphicsContext ;
  // iface->clearRect(5,5,20,20);
  unsigned int *screenBuffer = (unsigned int*) drawEvent->dstBuffer;
  // Draw a blue card - w00t
  // for(int i = 0 ;i < 320*480 ; i++) screenBuffer[i] = 0xaabbccdd ;
  InstanceData *data = (InstanceData *)instance->pdata;
  MyObject *myobj_ptr = data->object;
  if (!myobj_ptr) {
    return;
  }
  MyObject &myobj = *myobj_ptr;
  pthread_mutex_lock(&myobj.draw_mutex);
  myobj.checkSpawn();

  int dst_width = drawEvent->dstRight-drawEvent->dstLeft;
  int dst_height = drawEvent->dstBottom-drawEvent->dstTop;

#if 0
	const int screenWidth	= drawEvent->dstRowBytes / 4;
	int left	= max(0, drawEvent->dstLeft);
	int right	= min(screenWidth, drawEvent->dstRight);
	int top		= max(0, drawEvent->srcTop);
	int bottom	= drawEvent->srcBottom;
	debugLog(__LINE__, "HandleDrawEvent(sl:%i sr:%i st:%i sb:%i dl:%i dr:%i dt:%i db:%i drb:%i [top:%d left:%d bottom:%d right:%d screenWidth:%d] buf:%p myobj:%p)",
		drawEvent->srcLeft,
		drawEvent->srcRight,
		drawEvent->srcTop,
		drawEvent->srcBottom,
		drawEvent->dstLeft,
		drawEvent->dstRight,
		drawEvent->dstTop,
		drawEvent->dstBottom,
		drawEvent->dstRowBytes,
		top,
		left,
		bottom,
		right,
		screenWidth,
		screenBuffer,
		&myobj);
#endif

  FontInfo &font = myobj.char_buf.font();
  //debugLog(__LINE__, "font:%p", font);
  PixelBuffer pbuf(font, screenBuffer,dst_width,dst_height,drawEvent->dstRowBytes,drawEvent->srcTop);
  //debugLog(__LINE__, "pbuf:%p", &pbuf);
  pbuf.clear(PixelBuffer::pixel(0,0,0));

  {
    MyObject::CharBuf &char_buf = myobj.char_buf;
    const int width = myobj.char_buf.width();
    const int height = myobj.char_buf.height();
#if 0
    const int char_height = font.char_height();
    const int char_width = font.char_width();
    const int topRow = font.getLine(top);
    int row = topRow, y = top - (topRow * line_height);
    for (;row < height && y < bottom;++row, y += line_height) {
      int col = 0, x;
      for (x = left; col < width && x < right;++col, x += char_width) {
        pbuf.writeCharXY(x,y,char_buf.cell(row,col));
      }
    }
#else
    int begin_row = max(pbuf.topRow(),0);
    int end_row = min(pbuf.bottomRow(),height);
    {
      int row = begin_row;
      for (;row < end_row;++row) {
        int col = 0;
        for (;col<width;++col) {
          assert(row<char_buf.height());
          pbuf.writeChar(row,col,char_buf.cell(row,col));
        }
      }
    }
    assert(char_buf.currentRow()<char_buf.height());
    pbuf.invertChar(char_buf.currentRow(),char_buf.currentCol());
#endif
#if 0
    int cursor_y = cursor.row() * char_height;
    int cursor_x = cursor.col() * char_width;
    int add_y = (top % line_height);
    debugLog(__LINE__, "cursor.row():%d, cursor_y:%d, top:%d, drawEvent->dstTop:%d, topRow:%d, line_height:%d, add_y:%d (%d)", cursor.row(), cursor_y, top, drawEvent->dstTop, topRow, line_height, add_y, cursor_y + add_y);
    pbuf.invertCharXY(cursor_x, cursor_y + add_y - (add_y ? line_height : 0));
#endif
  }

  pthread_cond_signal(&myobj.drawn_cond);
  fprintf(stderr,"Done drawing\n");
  pthread_mutex_unlock(&myobj.draw_mutex);
  double end_time = CurrentTime();
  fprintf(stderr,"Drawing time: %g\n",end_time-start_time);
}


int16_t
NPP_HandleEvent(NPP instance, void* aEvent) {
  NpPalmEventType *palmEvent = (NpPalmEventType *) aEvent;
  switch (palmEvent->eventType) {
    case npPalmPenDownEvent:
    fprintf(stderr, "term: NPP_HandleEvent: npPalmPenDownEvent\n");
    break;
    case npPalmPenUpEvent:
    fprintf(stderr, "term: NPP_HandleEvent: npPalmPenUpEvent\n");
    break;
    case npPalmPenMoveEvent:
      fprintf(stderr, "term: NPP_HandleEvent: npPalmPenMoveEvent\n");
      HandlePenMoveEvent(&palmEvent->data.penEvent,instance);
    break;
    case npPalmKeyDownEvent:
      fprintf(stderr, "term: NPP_HandleEvent: npPalmKeyDownEvent\n");
      HandleKeyEvent(&palmEvent->data.keyEvent,instance);
    break;
    case npPalmKeyUpEvent:
    fprintf(stderr, "term: NPP_HandleEvent: npPalmKeyUpEvent\n");
    break;
    case npPalmKeyRepeatEvent:
    fprintf(stderr, "term: NPP_HandleEvent: npPalmKeyRepeatEvent\n");
    break;
    case npPalmKeyPressEvent:
    fprintf(stderr, "term: NPP_HandleEvent: npPalmKeyPressEvent\n");
    break;
    case npPalmDrawEvent:
      HandleDrawEvent(&palmEvent->data.drawEvent,instance);
      break;
    default:
    fprintf(stderr, "term: NPP_HandleEvent: unknown\n");
    break;
  }
  return 1;
}

void
NPP_URLNotify(NPP instance, const char* URL, NPReason reason, void* notifyData) {
  fprintf(stderr, "term: NPP_URLNotify\n");
}

const char * varname(NPPVariable variable)
{
  switch (variable) {
    case NPPVpluginNameString:
    return "NPPVpluginNameString";
    break;
    case NPPVpluginDescriptionString:
    return "NPPVpluginDescriptionString";
    break;
    case NPNVnetscapeWindow:
    return "NPNVnetscapeWindow";
    break;
    case NPPVpluginScriptableIID:
    return "NPPVpluginScriptableIID";
    break;
    case NPPVjavascriptPushCallerBool:
    return "NPPVjavascriptPushCallerBool";
    break;
    case NPPVpluginKeepLibraryInMemory:
    return "NPPVpluginKeepLibraryInMemory";
    break;
    case NPPVpluginScriptableNPObject:
    return "NPPVpluginScriptableNPObject";
    break;
    case NPPVpluginNeedsXEmbed:
    return "NPPVpluginNeedsXEmbed";
    break;
    case npPalmEventLoopValue:
    return "npPalmEventLoopValue";
    break;
    case npPalmCachePluginValue:
    return "npPalmCachePluginValue";
    break;
    case npPalmApplicationIdentifier:
    return "npPalmApplicationIdentifier";
    break;
    default:
    return "unknown";
    break;
  }
}

NPError
NPP_GetValue(NPP instance, NPPVariable variable, void *value) {
  logmsgv("NPP_GetValue",variable);
  switch (variable) {
    case NPPVpluginScriptableNPObject:
      *(NPObject **)value = GetScriptableObject(instance);
      break;
    default:
      return NPERR_GENERIC_ERROR;
  }
  return NPERR_NO_ERROR;
}

NPError
NPP_SetValue(NPP instance, NPNVariable variable, void *value) {
   fprintf(stderr, "term: NPP_SetValue");
  // logmsgv("NPP_SetValue",variable);
  return NPERR_GENERIC_ERROR;
}
